;=============================================================================
; Macros and other preprocessor constants
;=============================================================================

%MACRO cglobal 1
    %IFDEF NO_PREFIXED_SYMBOL
        global %1
    %ELSE
        global _%1
        %DEFINE %1 _%1
    %ENDIF
%ENDMACRO

%MACRO mt_mangle 2
   %IFDEF NO_PREFIXED_SYMBOL
      global %1_%2
   %ELSE
      global _%1_%2
      %DEFINE %2 _%1_%2
   %ENDIF
%ENDMACRO

;=============================================================================
; common rodata stuff
;=============================================================================

SECTION .rodata ALIGN=16
ALIGN 16
   MT_MASK_LEFT:     DB 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
   MT_MASK_RIGHT:    DB 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF

;=============================================================================
; Macros
;=============================================================================

%DEFINE PARAM(x)    EQU _MT_STACK * 4 + x * 4 + _MT_PUSH_COUNT * 4
%DEFINE PSTACK(x)   EQU x * 4

%MACRO MT_PUSH 1-*
    %REP %0
        push        %1
        %ROTATE     1
    %ENDREP
%ENDMACRO

%MACRO MT_POP 1-*
    %REP %0
        %ROTATE     -1
        pop         %1
    %ENDREP
%ENDMACRO

%MACRO MT_SET_STACK 1-*
    %ASSIGN _MT_STACK %1
    %ASSIGN _MT_PUSH_COUNT %0 - 1
    %XDEFINE        MT_REGISTERS
    %IF (%0 > 1)
        %XDEFINE    MT_REGISTERS %2
    %ENDIF
    %IF (%0 > 2)
        %XDEFINE    MT_REGISTERS MT_REGISTERS, %3
    %ENDIF
    %IF (%0 > 3)
        %XDEFINE    MT_REGISTERS MT_REGISTERS, %4
    %ENDIF
    %IF (%0 > 4)
        %XDEFINE    MT_REGISTERS MT_REGISTERS, %5
    %ENDIF
    %IF (%0 > 1)
        MT_PUSH     MT_REGISTERS
    %ENDIF
    %IF (%1 > 0)
        sub         esp, _MT_STACK * 4
    %ENDIF
%ENDMACRO

%MACRO MT_REL_STACK 0-1
    %IF (%0 == 0)
         %IF (_MT_STACK > 0)
            add         esp, _MT_STACK * 4
         %ENDIF
    %ENDIF
    MT_POP        MT_REGISTERS
    %UNDEF          MT_REGISTERS
    ret
%ENDMACRO

%MACRO PUSHALL 0
	push      ebx
	push      edi
	push      esi
	push      ebp
%ENDMACRO

%MACRO POPALL 0
	pop       ebp
	pop       esi
	pop       edi
	pop       ebx
%ENDMACRO

%DEFINE PARAMETER(n)     EQU 16 + 4 * n

%MACRO SHUFFLE 6
    pshufw          %1, %2, %3 * 1 + %4 * 4 + %5 * 16 + %6 * 64
%ENDMACRO


;=============================================================================
; prefetchs
;=============================================================================

%MACRO MT_PREFETCHW  1
   prefetchw         [%1]
%ENDMACRO

%MACRO MT_PREFETCHT0 1
   prefetcht0        [%1]
%ENDMACRO

%MACRO MT_NOPREFETCH 1
%ENDMACRO

%MACRO MT_PREFETCHNTA 1
   prefetchnta       [%1]
%ENDMACRO


;=============================================================================
; unrolling stuff
;=============================================================================


; uses edx, eax, ecx, edi
; edi is the pointer on the data, eax contains the pitch, and ecx the height
; three parameters : code for the aligning loops ( tail - 8 bytes)
;                    code for the unrolled loop ( 64 bytes )
;                    size of the jump in the unrolled loop
; in these loops, edi + eax will be the current position
%MACRO MT_UNROLL_INPLACE 2
   mov                  eax, [esp + .nDstPitch]
   mov                  ecx, [esp + .nHeight]

   mul                  ecx               ; compute the total size
   sub                  eax, ebp
   
   mov                  edx, eax          ; save it into edx ( unrolled loop stop )
   mov                  ecx, eax          ; save it into ecx
   and                  edx, 0xFFFFFFC0   ; make the loop 64-bytes aligned
   xor                  eax, eax          ; reset eax ( loop counter )
   
   cmp                  eax, ecx
   je                   .end              ; if nothing, jump
   
.loop

   %1                                     ; unrolled code
   
   add                  eax, 64
   cmp                  eax, edx
   jl                   .loop
   
   cmp                  eax, ecx
   je                   .end
   
.tail

   %2                                     ; tail code

   add                  eax, 8
   cmp                  eax, ecx
   jl                   .tail
   
.end
   
%ENDMACRO   


; uses edx, eax, ecx
; ecx will be the height counter, eax the width one, and edx the loop stop
; three parameters : code for the unrolled loop ( 64 bytes )
;                    code for the aligning loops ( head & tail - 8 bytes)
;                    code for increasing pointers
%MACRO MT_UNROLL_INPLACE_WIDTH 3
   mov                  ecx, [esp + .nHeight];
   mov                  edx, [esp + .nWidth] ; save it into edx ( unrolled loop stop )
   and                  edx, 0xFFFFFFC0      ; make the loop 64-bytes aligned
   
%%loopy
   
   xor                  eax, eax          ; reset eax ( loop counter )
   
   cmp                  eax, edx          ; check that we can do 64 pixels 
   je                   %%no64
   
%%loop

   %1                                     ; unrolled code
   
   add                  eax, 64
   cmp                  eax, edx
   jl                   %%loop
   
%%no64   
   
   cmp                  eax, [esp + .nWidth]
   je                   %%end

%%tail

   %2                                     ; tail code ( same as head )

   add                  eax, 8
   cmp                  eax, [esp + .nWidth]
   jl                   %%tail
   
%%end

   %3                                     ; change line code
   dec                  ecx
   jnz                  %%loopy


   
%ENDMACRO   
   
;=============================================================================
; convolution & border stuff
;=============================================================================

%MACRO MT_LOAD_NORMAL      6
   %1                      %2, [%3 + %4]
%ENDMACRO

%MACRO MT_LOAD_OFFSET      7
   %IFNIDN %1, movq
      movq                 %3, [%5]
      movq                 %4, [%6]
      pand                 %4, %3
      %7                   %3, 8          ; drop the byte
      paddb                %3, %4
      %1                   %2, %3
   %ELSE
      movq                 %2, [%5]
      movq                 %4, [%6]
      pand                 %4, %2
      %7                   %2, 8          ; drop the byte
      paddb                %2, %4
   %ENDIF
%ENDMACRO

   
;=============================================================================
;left border
;=============================================================================

%MACRO MT_LOAD_LEFT_BORDER    6
   MT_LOAD_OFFSET       %1, %2, %5, %6, %3 + %4, MT_MASK_LEFT, psllq
%ENDMACRO

%MACRO MT_LOAD_LEFT_NORMAL    6
   %1                   %2, [%3 + %4 - 1]
%ENDMACRO

;=============================================================================
;right border
;=============================================================================

%MACRO MT_LOAD_RIGHT_BORDER   6
   MT_LOAD_OFFSET       %1, %2, %5, %6, %3 + %4, MT_MASK_RIGHT, psrlq
%ENDMACRO

%MACRO MT_LOAD_RIGHT_NORMAL   6
   %1                   %2, [%3 + %4 + 1]
%ENDMACRO

;=============================================================================
; top border
;=============================================================================

%MACRO MT_LOAD_UP_BORDER      6
   %1                   %2, [%3 + %4]
%ENDMACRO

%MACRO MT_LOAD_UP_NORMAL      6
   %1                   %2, [%3]
%ENDMACRO

;=============================================================================
; bottom border
;=============================================================================

%MACRO MT_LOAD_DOWN_BORDER    6
   %1                   %2, [%3 + %4]
%ENDMACRO

%MACRO MT_LOAD_DOWN_NORMAL    6
   %1                   %2, [%3 + 2 * %4]
%ENDMACRO

;=============================================================================
; top left
;=============================================================================

%MACRO MT_LOAD_UP_LEFT_BORDER_UP_LEFT 6
   MT_LOAD_OFFSET       %1, %2, %5, %6, %3 + %4, MT_MASK_LEFT, psllq
%ENDMACRO

%MACRO MT_LOAD_UP_LEFT_BORDER_LEFT 6
   MT_LOAD_OFFSET       %1, %2, %5, %6, %3, MT_MASK_LEFT, psllq
%ENDMACRO

%MACRO MT_LOAD_UP_LEFT_BORDER_UP 6
   %1                   %2, [%3 + %4 - 1]
%ENDMACRO

%MACRO MT_LOAD_UP_LEFT_NORMAL 6
   %1                   %2, [%3 - 1]
%ENDMACRO

;=============================================================================
; top right
;=============================================================================

%MACRO MT_LOAD_UP_RIGHT_BORDER_UP_RIGHT 6
   MT_LOAD_OFFSET       %1, %2, %5, %6, %3 + %4, MT_MASK_RIGHT, psrlq
%ENDMACRO

%MACRO MT_LOAD_UP_RIGHT_BORDER_RIGHT 6
   MT_LOAD_OFFSET       %1, %2, %5, %6, %3, MT_MASK_RIGHT, psrlq
%ENDMACRO

%MACRO MT_LOAD_UP_RIGHT_BORDER_UP 6
   %1                   %2, [%3 + %4 + 1]
%ENDMACRO

%MACRO MT_LOAD_UP_RIGHT_NORMAL 6
   %1                   %2, [%3 + 1]
%ENDMACRO

;=============================================================================
; bottom left
;=============================================================================

%MACRO MT_LOAD_DOWN_LEFT_BORDER_LEFT 6
   MT_LOAD_OFFSET       %1, %2, %5, %6, %3 + 2 * %4, MT_MASK_LEFT, psllq
%ENDMACRO

%MACRO MT_LOAD_DOWN_LEFT_BORDER_DOWN_LEFT 6
   MT_LOAD_OFFSET       %1, %2, %5, %6, %3 + %4, MT_MASK_LEFT, psllq
%ENDMACRO

%MACRO MT_LOAD_DOWN_LEFT_BORDER_DOWN 6
   %1                   %2, [%3 + %4 - 1]
%ENDMACRO

%MACRO MT_LOAD_DOWN_LEFT_NORMAL 6
   %1                   %2, [%3 + 2 * %4 - 1]
%ENDMACRO

;=============================================================================
; bottom right
;=============================================================================

%MACRO MT_LOAD_DOWN_RIGHT_BORDER_RIGHT 6
   MT_LOAD_OFFSET       %1, %2, %5, %6, %3 + 2 * %4, MT_MASK_RIGHT, psrlq
%ENDMACRO

%MACRO MT_LOAD_DOWN_RIGHT_BORDER_DOWN_RIGHT 6
   MT_LOAD_OFFSET       %1, %2, %5, %6, %3 + %4, MT_MASK_RIGHT, psrlq
%ENDMACRO

%MACRO MT_LOAD_DOWN_RIGHT_BORDER_DOWN 6
   %1                   %2, [%3 + %4 + 1]
%ENDMACRO

%MACRO MT_LOAD_DOWN_RIGHT_NORMAL 6
   %1                   %2, [%3 + 2 * %4 + 1]
%ENDMACRO

; combos

%DEFINE MT_TOP_LEFT     MT_LOAD_UP_LEFT_BORDER_UP_LEFT, MT_LOAD_UP_BORDER, MT_LOAD_UP_RIGHT_BORDER_UP      , MT_LOAD_LEFT_BORDER, MT_LOAD_RIGHT_NORMAL, MT_LOAD_DOWN_LEFT_BORDER_LEFT     , MT_LOAD_DOWN_NORMAL, MT_LOAD_DOWN_RIGHT_NORMAL
%DEFINE MT_TOP          MT_LOAD_UP_LEFT_BORDER_UP     , MT_LOAD_UP_BORDER, MT_LOAD_UP_RIGHT_BORDER_UP      , MT_LOAD_LEFT_NORMAL, MT_LOAD_RIGHT_NORMAL, MT_LOAD_DOWN_LEFT_NORMAL          , MT_LOAD_DOWN_NORMAL, MT_LOAD_DOWN_RIGHT_NORMAL
%DEFINE MT_TOP_RIGHT    MT_LOAD_UP_LEFT_BORDER_UP     , MT_LOAD_UP_BORDER, MT_LOAD_UP_RIGHT_BORDER_UP_RIGHT, MT_LOAD_LEFT_NORMAL, MT_LOAD_RIGHT_BORDER, MT_LOAD_DOWN_LEFT_NORMAL          , MT_LOAD_DOWN_NORMAL, MT_LOAD_DOWN_RIGHT_BORDER_RIGHT

%DEFINE MT_LEFT         MT_LOAD_UP_LEFT_BORDER_LEFT   , MT_LOAD_UP_NORMAL, MT_LOAD_UP_RIGHT_NORMAL         , MT_LOAD_LEFT_BORDER, MT_LOAD_RIGHT_NORMAL, MT_LOAD_DOWN_LEFT_BORDER_LEFT     , MT_LOAD_DOWN_NORMAL, MT_LOAD_DOWN_RIGHT_NORMAL
%DEFINE MT_NORMAL       MT_LOAD_UP_LEFT_NORMAL        , MT_LOAD_UP_NORMAL, MT_LOAD_UP_RIGHT_NORMAL         , MT_LOAD_LEFT_NORMAL, MT_LOAD_RIGHT_NORMAL, MT_LOAD_DOWN_LEFT_NORMAL          , MT_LOAD_DOWN_NORMAL, MT_LOAD_DOWN_RIGHT_NORMAL
%DEFINE MT_RIGHT        MT_LOAD_UP_LEFT_NORMAL        , MT_LOAD_UP_NORMAL, MT_LOAD_UP_RIGHT_BORDER_RIGHT   , MT_LOAD_LEFT_NORMAL, MT_LOAD_RIGHT_BORDER, MT_LOAD_DOWN_LEFT_NORMAL          , MT_LOAD_DOWN_NORMAL, MT_LOAD_DOWN_RIGHT_BORDER_RIGHT

%DEFINE MT_BOTTOM_LEFT  MT_LOAD_UP_LEFT_BORDER_LEFT   , MT_LOAD_UP_NORMAL, MT_LOAD_UP_RIGHT_NORMAL         , MT_LOAD_LEFT_BORDER, MT_LOAD_RIGHT_NORMAL, MT_LOAD_DOWN_LEFT_BORDER_DOWN_LEFT, MT_LOAD_DOWN_BORDER, MT_LOAD_DOWN_RIGHT_BORDER_DOWN
%DEFINE MT_BOTTOM       MT_LOAD_UP_LEFT_NORMAL        , MT_LOAD_UP_NORMAL, MT_LOAD_UP_RIGHT_NORMAL         , MT_LOAD_LEFT_NORMAL, MT_LOAD_RIGHT_NORMAL, MT_LOAD_DOWN_LEFT_BORDER_DOWN     , MT_LOAD_DOWN_BORDER, MT_LOAD_DOWN_RIGHT_BORDER_DOWN
%DEFINE MT_BOTTOM_RIGHT MT_LOAD_UP_LEFT_NORMAL        , MT_LOAD_UP_NORMAL, MT_LOAD_UP_RIGHT_BORDER_RIGHT   , MT_LOAD_LEFT_NORMAL, MT_LOAD_RIGHT_BORDER, MT_LOAD_DOWN_LEFT_BORDER_DOWN     , MT_LOAD_DOWN_BORDER, MT_LOAD_DOWN_RIGHT_BORDER_DOWN_RIGHT

;=============================================================================
; processing
;=============================================================================

%MACRO MT_PROCESS_LINE     7
   xor                     eax, eax                         ; loop counter resetted
   
   ; left
   %2                      %3                               ; rolled code, left
   
   add                     eax, 8
   %6                                                       ; next 8 pixels
   
%%loopx:

   %1                      %4                               ; unrolled code, center
   add                     eax, 64
   %7                                                       ; next 64 pixels
   cmp                     eax, [esp + .nUnrolledWidth]
   jl                      %%loopx
   
   cmp                     eax, [esp + .nTailWidth]
   je                      %%endx
   
%%tailx:

   %2                      %4                               ; rolled code, center
   add                     eax, 8
   %6                                                       ; next 8 pixels
   cmp                     eax, [esp + .nTailWidth]
   jl                      %%tailx
  
%%endx:   
   ; right
   
   %2                      %5                               ; rolled code, right
   %6                                                       ; next 8 pixels
%ENDMACRO   
   
%MACRO MT_PROCESS_CONVOLUTION 5
   mov                     eax, [esp + .nWidth]
   sub                     eax, 8
   mov                     [esp + .nTailWidth], eax         ; row length, minus 8 last bytes
   sub                     eax, 8                           ; row length, minus 8 first bytes
   and                     eax, 0xFFFFFFC0                  ; multiple of 64
   add                     eax, 8                           ; add back the 8 first bytes
   mov                     [esp + .nUnrolledWidth], eax     ; row length, rounded to the last unrolled loop

   ; top line

   MT_PROCESS_LINE         %1, %2, {MT_TOP_LEFT}, {MT_TOP}, {MT_TOP_RIGHT}, %4, %5
   
   %3                                                       ; next line
   
   ; center
   mov                     eax, [esp + .nHeight]
   sub                     eax, 2                           ; -2 : top & bottom lines
   
.loopy

   mov                     [esp + .nHeightCount], eax       ; save height counter

   MT_PROCESS_LINE         %1, %2, {MT_LEFT}, {MT_NORMAL}, {MT_RIGHT}, %4, %5

   %3                                                       ; next line
   
   mov                     eax, [esp + .nHeightCount]             ; load height counter
   dec                     eax
   jnz                     .loopy
   
   ; bottom line
   
   MT_PROCESS_LINE         %1, %2, {MT_BOTTOM_LEFT}, {MT_BOTTOM}, {MT_BOTTOM_RIGHT}, %4, %5
%ENDMACRO